import sys
from io import StringIO
from re import match

from pymortafix.utils import multisub, strict_input

# --- utility


def force_input(inputs):
    sys.stdin = StringIO(inputs)


# --- multisub


def test_multisub_no_seq_no_overlap():
    string = "Test 42"
    sub_dict = {}
    assert multisub(sub_dict, string) == "Test 42"
    sub_dict = {"st": "ST"}
    assert multisub(sub_dict, string) == "TeST 42"
    sub_dict = {r"\s": "_", r"\d+": "@"}
    assert multisub(sub_dict, string) == "Test_@"


def test_multisub_no_seq_overlap():
    string = "Test - Overlap"
    sub_dict = {"-": " ", r"\s+": "_"}
    assert multisub(sub_dict, string) == "Test_ _Overlap"
    sub_dict = {r"\s+": "_", "-": " "}
    assert multisub(sub_dict, string) == "Test_ _Overlap"


def test_multisub_seq():
    string = "Test -  Sequential  - Not  Bad"
    sub_dict = {r"-": ":", r"\s+:\s+": "-", r"\s+": "_"}
    assert multisub(sub_dict, string, sequential=True) == "Test-Sequential-Not_Bad"


# --- strict input


def test_strictinput_easy():
    force_input("test")
    inp = strict_input("Input: ")
    assert inp == "test"


def test_strictinput_choices():
    force_input("test\nup")
    inp = strict_input("Input: ", choices=("up", "down"))
    assert inp == "up"
    force_input("left\ndown\nup")
    inp = strict_input("Input: ", choices=("up", "down"))
    assert inp == "down"


def test_strictinput_regex():
    force_input("test-2\ntest-42")
    inp = strict_input("Input: ", regex=r"test-\d+")
    assert inp == "test-2"
    force_input("test\ntest-42")
    inp = strict_input("Input: ", regex=r"test-\d+")
    assert inp == "test-42"


def test_strictinput_check():
    force_input("uno\n1\n42")
    inp = strict_input("Input: ", check=lambda x: match(r"\d+$", x) and int(x) % 2 == 0)
    assert inp == "42"
    force_input("1\n-10\n42")
    inp = strict_input("Input: ", check=lambda x: 2 ** int(x) > 1000)
    assert inp == "42"


def test_strictinput_choices_regex():
    force_input("test\nup\ntest-42")
    inp = strict_input("Input: ", choices=("up", "down"), regex=r"test-\d+")
    assert inp == "up"
    force_input("test\nleft\ntest-42")
    inp = strict_input("Input: ", choices=("up", "down"), regex=r"test-\d+")
    assert inp == "test-42"


def test_strictinput_choices_check():
    force_input("23\n3\n42")
    inp = strict_input("Input: ", choices=("3"), check=lambda x: int(x) % 2 == 0)
    assert inp == "3"
    force_input("23\n5\n42")
    inp = strict_input("Input: ", choices=("3"), check=lambda x: int(x) % 2 == 0)
    assert inp == "42"


def test_strictinput_regex_check():
    force_input("23\nu911\n42")
    inp = strict_input("Input: ", regex=r"u\d+", check=lambda x: int(x) % 2 == 0)
    assert inp == "u911"
    force_input("23\n911\n42")
    inp = strict_input("Input: ", regex=r"u\d+", check=lambda x: int(x) % 2 == 0)
    assert inp == "42"



from re import findall, match, sub

from colorifix.colorifix import erase
from pymortafix._getchar import _Getch


def get_sub_from_matching(dictionary, matching):
    index = [i for i, group in enumerate(matching.groups()) if group]
    matched = list(dictionary)[index[0]] if index else None
    return dictionary.get(matched)


def multisub(sub_dict, string, sequential=False):
    """Infinite sub in one iteration # sub_dict: {what_to_sub:substitution}"""
    if not sequential:
        rgx = "|".join(f"({s})" for s in sub_dict.keys())
        return sub(rgx, lambda m: get_sub_from_matching(sub_dict, m), string)
    else:
        for rgx, substitution in sub_dict.items():
            string = sub(rgx, substitution, string)
        return string


def strict_input(
    text, wrong_text=None, choices=None, regex=None, check=None, flush=False
):
    """Get user input with some requirements"""
    inp = input(text)
    if flush:
        erase(len(findall(r"\n", text)) + 1)
    while (
        (not choices or choices and inp not in choices)
        and (not regex or regex and not match(regex, inp))
        and (not check or check and not check(inp))
        and (choices or regex or check)
    ):
        if wrong_text:
            inp = input(wrong_text)
        else:
            inp = input(text)
        if flush:
            erase(len(findall(r"\n", wrong_text or text)) + 1)
    return inp


def direct_input(choices=None):
    """Get user single char input w/o return, with optional restricted choices"""
    inkey = _Getch()
    k = inkey()
    while (
        choices
        and (isinstance(k, str) and k or k.decode(errors="replace")) not in choices
    ):
        k = inkey()
    return isinstance(k, str) and k or k.decode(errors="replace")



import sys
import traceback
from datetime import datetime

from colorifix.colorifix import paint
from emoji import emojize

# ---- Utils


def emz(text):
    """emojize everything"""
    if isinstance(text, list):
        return [emz(t) for t in text]
    if isinstance(text, dict):
        return {k: emz(v) for k, v in text.items()}
    return emojize(text, variant="emoji_type")


# ---- Messages


def standard_message(update, top_message, bot_name, bot_url):
    """Build a standard log message"""
    user_name = update.effective_chat.username or update.effective_chat.first_name
    user_id = update.effective_chat.id
    user_mention = f"[{user_name}](tg://user?id={user_id})"
    now = datetime.now()
    return emojize(
        f"{top_message}\n"
        f":robot: [{bot_name}]({bot_url})\n"
        f":calendar: {now:%d.%m.%Y}\n"
        f":eight_o’clock: {now:%H:%M}\n"
        f":bust_in_silhouette: {user_mention} (`{user_id}`)\n"
    )


def send_log(bot, channel, message, document=None):
    """Send a simple message to a channel"""
    if document:
        bot.send_document(
            channel,
            document,
            caption=emojize(message),
            parse_mode="Markdown",
        )
    else:
        bot.send_message(channel, emojize(message), parse_mode="Markdown")


def error_handler(
    update,
    context,
    channel,
    message,
    logger=None,
    exception=None,
    extra_info=None,
    log=True,
):
    """Log an error on terminal and a Telegram channel from a Telegram bot"""
    # error log on terminal
    user_data = context.user_data if update else "None"
    trace = (exception and exception.__traceback__) or sys.exc_info()[2]
    traceback_error = "".join(traceback.format_tb(trace))
    exp = exception or context.error
    traceback_msg = paint(
        f"[#229]{user_data}\n\n[#gray]{traceback_error}" f"[#red]{exp}", False
    )
    if logger:
        logger.warning(traceback_msg)
    else:
        print(traceback_msg)
    # send message to error channel
    if log:
        open("error.log", "w+").write(
            f"{user_data}\n\n"
            f"{extra_info or '...'}\n\n"
            f"{traceback_error}\n"
            f"> {exp}"
        )
        send_log(context.bot, channel, message, open("error.log", "r"))


def not_authorized_handler(update, context, bot_name, bot_url, channel):
    """Log an unauthorized access on a Telegram bot"""
    top_msg = ":face_with_symbols_on_mouth: *NO AUTH USER* :face_with_symbols_on_mouth:"
    message = standard_message(update, top_msg, bot_name, bot_url)
    send_log(context.bot, channel, message)


def sentry_log(project, sentry_url, user):
    """send a telegram log for Sentry"""
    msg = emz(
        f":construction: *{project}* :construction:\n\n"
        f":calendar: {datetime.now():%d.%m.%Y}\n"
        f":bust_in_silhouette: *{user}*\n"
        f":microbe: [Sentry Issue]({sentry_url})"
    )
    return msg
